/**
 * @brief Interface Driver for SD cards
 * @author Michael Burmeister
 * @date Feburary 1, 2021
 * @version 1.0
 * 
*/

//#define _DEBUG

#include <stdio.h>
#include <propeller.h>
#include "sd_mmc.h"


static int _CS = PIN_SS;
static int _CLK = PIN_CLK;
static int _MOSI = PIN_MOSI;
static int _MISO = PIN_MISO;


void SetSD(int cs, int clk, int mosi, int miso)
{
	_CS = cs;
	_CLK = clk;
	_MOSI = mosi;
	_MISO = miso;
}


void EnableSD()
{
    _waitms(10);
    _dirh(_CS);
    _pinh(_CS);
    _dirh(_CLK);
    _pinl(_CLK);
    _dirh(_MOSI);
    _dirl(_MISO);
}

void SendSD(const char* buff, unsigned int bc)
{
    int i;
    char *b;
    char x;
    int p = _MOSI;
    int pp = _CLK;
    int j;
    int t = bc;

    b = buff;
    __asm volatile {
            mov   i, t
    loopi   mov   j, #8
            rdbyte x, b
    loopj   test  x, #0x80  wz
      if_z  drvl p
      if_nz drvh p
            drvh pp
            drvl pp
            rol x, #1
            djnz j, #loopj
            add b, #1
            djnz i, #loopi
      }
}

void ReceiveSD(char *buff, unsigned int bc)
{
    char *b = buff;
    unsigned int c = bc;
    int m = _MOSI;
    int n = _MISO;
    int x = _CLK;
    int i;
    int j;
    int v;

    __asm volatile {
           drvh m
           waitx #8
           mov i, c
    loopi  mov v, #0
           mov j, #8
    loopj  rol v, #1
           testp n wc
      if_c add v, #1
           drvh x
           drvl x
           waitx #8
           djnz j, #loopj
           wrbyte v, b
           add b, #1
           djnz i, #loopi
    }
}

int GetStatus(void)
{
    int i;
    char b;
    
    for (i=0;i<500;i++)
    {
        ReceiveSD(&b, 1);
        if (b == 0xff)
        	return 1;
        _waitms(1);
    }
	return 0;
}

void ReleaseSD(void)
{
    char b;
    
    _pinh(_CS);
    ReceiveSD(&b, 1);
}

int SelectSD(void)
{
    char b;
    
    _pinl(_CS);
    ReceiveSD(&b, 1);
    if (GetStatus())
    	return 1;
    ReleaseSD();
    return 0;
}

int ReceiveBlock(char *buff, unsigned int bc)
{
    char b[2];
    int i;
    
    for (i=0;i<100;i++)
    {
        ReceiveSD(b, 1);
        if (*b != 0xff)
        	break;
        _waitms(1);
    }
	if (*b != 0xfe)
		return 0;
	ReceiveSD(buff, bc);
	ReceiveSD(b, 2);    // Ditch CRC
	return 1;
}

int SendBlock(char *buff, char token)
{
    char b[2];
    
    if (GetStatus() == 0)
    	return 0;
    
    b[0] = token;
    SendSD(b, 1);
    if (token != 0xfd)
    {
        SendSD(buff, 512);
        ReceiveSD(b, 2);  // CRC value
        ReceiveSD(b, 1);
        if ((*b & 0x1f) != 0x05)
        	return 0;
    }
	return 1;    
}

char SendCommand(char cmd, unsigned int arg)
{
    char b[6];
    char n;
    
    if (cmd & 0x80)
    {
        cmd = cmd & 0x7f;
        n = SendCommand(CMD55, 0);
        if (n > 1)
        	return n;
    }
    
    if (cmd != CMD12)
    {
        ReleaseSD();
        if (!SelectSD())
        	return 0xff;
    }
    
    b[0] = 0x40 | cmd;
    b[1] = arg >> 24;
    b[2] = arg >> 16;
    b[3] = arg >> 8;
    b[4] = arg;
    n = 0x01;
    if (cmd == CMD0)
    	n = 0x95;
    if (cmd == CMD8)
    	n = 0x87;
    b[5] = n;
    SendSD(b, 6);
    
    if (cmd == CMD12)
    	ReceiveSD(&n, 1);

    for (int i=0;i<10;i++)
    {
        ReceiveSD(&n, 1);
        if ((n & 0x80) == 0)
        	break;
    }
#ifdef _DEBUG
    __builtin_printf("Command: %d (%x) --> %d\n", cmd, arg, n);
#endif
    return n;
}
